#!/usr/bin/perl

# -----------------------------------------------------------------------------
# Name: aimChomp (aimchomp.pl)
# Auth: Jason Bittel
# Date: 04/23/04
# Desc: Parser for aimSniff text dump files. Takes an aimSniff text file as
#       input and converts it to human readable format, while filtering by
#       any desired regexp.  Run the script with the -h option for all options.
#
# Todo: Add sanity checking on args; allow any text file to be used as a
#       hitlist-- -l designates input file and -o used for output
#
# -----------------------------------------------------------------------------
use strict;
use Getopt::Std;

# -----------------------------------------------------------------------------
# GLOBAL CONSTANTS
# -----------------------------------------------------------------------------
my $PATTERN = "##";             # Delimeter pattern for fields
my $HITLIST = "textfilter.dat"; # File containing hitlist items

# -----------------------------------------------------------------------------
# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
my %opts;          # Command line arguments
my $filter_regexp; # Regular expression filter applied to incoming data
my $filter_day;    # Day filter for incoming records
my $filter_month;  # Month filter for incoming records
my $filter_year;   # Year filter for incoming records
my @input_files;   # List of input files to read data from
my $output_file;   # Output file to pipe results to (optional)
my $hitlist_file;  # Output file for hitlist results (optional)
my $message_type;  # Message type for each record (optional)

# -----------------------------------------------------------------------------
# Main Program
# -----------------------------------------------------------------------------
&get_arguments();

# Compare text against hitlist and output results
if ($hitlist_file) {
   if (-e $HITLIST) {
      &scan_engine($hitlist_file, &parse_hitlist($HITLIST), $message_type, @input_files);
   } else {
      print "\nError: cannot find $HITLIST\n";
      print "\nHitlist not found...ignoring -l option\n";
   }
}

# Output parsed text lines
if ($output_file) {
   &scan_engine($output_file, $filter_regexp, $message_type, @input_files);
}

# -----------------------------------------------------------------------------
# Core engine, parses all input file based on options provided 
# -----------------------------------------------------------------------------
sub scan_engine {
   my $output_file   = shift;      # Output file for results
   my $filter_regexp = shift;      # Regular expression run on each line
   my $filter_family = shift;      # Family of messages to filter by
   my @input_files   = @_;         # Input files to parse

   my $curr_line; # Current line in input file
   my $curr_file; # Current input file
   my $output;    # Holds reference to output stream
   my @record;    # Entire record line
   my %fields;    # Record broken into fields

   # Create a reference to output medium
   if ($output_file) {
      open(OUTFILE, ">$output_file") || die("\nError: Cannot open $output_file - $!\n");
      $output = *OUTFILE{IO};
   } else {
      $output = *STDOUT{IO};
   }
   
   foreach $curr_file (@input_files) {
      unless(open(INFILE, "$curr_file")) {
         print "\nError: Cannot open $curr_file - $!\n";
         next;
      }
      
      foreach $curr_line (<INFILE>) {
         # Regexp filter (defaults to accepting everything)
         if ($curr_line =~ /$filter_regexp/i) {
            &clean_text(\$curr_line);

            # Parse record line into hash
            @record = split(/$PATTERN/, $curr_line);
            for (my $i = 1; $i < scalar @record; $i += 2) {
               $fields{$record[$i]} = $record[$i + 1];
            }

            # Check message family (optional)
            if ($filter_family) {
               next unless ($fields{'FAMILY'} =~ /$filter_family/);
            }
            
            # Check date match (optional)
            if ($filter_day || $filter_month || $filter_year) {
               next unless &filter_date(\$fields{'TS'}, $filter_day, $filter_month, $filter_year);
            }

            # Clean up record data
            &format_timestamp(\$fields{'TS'});
            if ($fields{'FROM'} eq "") { $fields{'FROM'} = "%from_unkn%"; }
            if ($fields{'DESTHANDLE'} eq "") { $fields{'DESTHANDLE'} = "%dest_unkn%"; }

            # Output line information to output handle
            if ($fields{'FAMILY'} eq "00040006") { # Outgoing Message
               print $output "$fields{'TS'}\t$fields{'FROM'}\t$fields{'DESTHANDLE'}\tS$fields{'SRCIP'}\t$fields{'MESSAGE'}\n";
            } elsif ($fields{'FAMILY'} eq "00040007") { # Incoming Message
               print $output "$fields{'TS'}\t$fields{'FROM'}\t$fields{'DESTHANDLE'}\tD$fields{'DESTIP'}\t$fields{'MESSAGE'}\n";
            } elsif ($fields{'FAMILY'} eq "00020015") { # Buddy Info Request
               print $output "$fields{'TS'}\t$fields{'BUDDY'}\tS$fields{'SRCIP'}\t$fields{'TYPE'}\n";
            } elsif ($fields{'FAMILY'} eq "00170006") { # Login
               print $output "$fields{'TS'}\t$fields{'HANDLE'}\tS$fields{'SRCIP'}\t$fields{'TYPE'}\n";
            } elsif ($fields{'FAMILY'} eq "00130006") { # Buddy List
               print $output "$fields{'TS'}\tD$fields{'DSTIP'}\t$fields{'TYPE'}\n";
            } else {
               print $output "$fields{'TS'}\t$fields{'TYPE'}\n";
            }
         }
      }
      close(INFILE);
   }

   if ($output_file) { close(OUTFILE); }
}

# -----------------------------------------------------------------------------
# Parse hitlist file, return regular expression string
# -----------------------------------------------------------------------------
sub parse_hitlist {
   my $hitlist = shift;
   my $hitlist_filter = "";

   open(HITLIST, "$hitlist") || die("\nError: Cannot open $hitlist - $!\n");
   while (<HITLIST>) {
      chomp;
      if (!$hitlist_filter) {
         $hitlist_filter .= "$_";
      } else {
         $hitlist_filter .= "|$_";
      }
   }
   close(HITLIST);

   return $hitlist_filter;
}

# -----------------------------------------------------------------------------
# Filter output messages based on date fields
# -----------------------------------------------------------------------------
sub filter_date {
   my ($year, $month, $day, $hour, $min, $sec);
   my $ts           = shift;
   my $filter_day   = shift;
   my $filter_month = shift;
   my $filter_year  = shift;
   my $op;

   # Break date into respective fields
   $$ts =~ /^(\d*)-(\d*)-(\d*)/;
   ($year, $month, $day) = ($1, $2, $3);
   
   if ($filter_day) {
      $filter_day =~ /^([\+\=\-])(\d+)/;
      ($op, $filter_day) = ($1, $2);

      if ($op eq '+') {
         return 1 if ($day > $filter_day);
      }
      if ($op eq '=') {
         return 1 if ($day == $filter_day);
      }
      if ($op eq '-') {
         return 1 if ($day < $filter_day);
      }
   }
   
   if ($filter_month) {
      $filter_month =~ /^([\+\=\-])(\d+)/;
      ($op, $filter_month) = ($1, $2);

      if ($op eq '+') {
         return 1 if ($month > $filter_month);
      }
      if ($op eq '=') {
         return 1 if ($month == $filter_month);
      }
      if ($op eq '-') {
         return 1 if ($month < $filter_month);
      }
   }
   
   if ($filter_year) {
      $filter_year =~ /^([\+\=\-])(\d+)/;
      ($op, $filter_year) = ($1, $2);

      if ($op eq '+') {
         return 1 if ($year > $filter_year);
      }
      if ($op eq '=') {
         return 1 if ($year == $filter_year);
      }
      if ($op eq '-') {
         return 1 if ($year < $filter_year);
      }
   }
   
   return 0;
}

# -----------------------------------------------------------------------------
# Clean up the date/time stamp
# -----------------------------------------------------------------------------
sub format_timestamp {
   my ($year, $month, $day, $hour, $min, $sec);
   my $ts = shift;

   # Input:  "2004-4-27 14:57:3"
   # Output: "4-27-2004 14:57:03"
   $$ts =~ /(\d*)-(\d*)-(\d*) (\d*):(\d*):(\d*)/;
   ($year, $month, $day, $hour, $min, $sec) = ($1, $2, $3, $4, $5, $6);

   $year = substr($year, 2, 2);

   if ($hour < 10) {
      $hour = "0" . $hour;
   }
   if ($min < 10) {
      $min = "0" . $min;
   }
   if ($sec < 10) {
      $sec = "0" . $sec;
   }

   $$ts = "$month-$day-$year $hour:$min:$sec";
}

# -----------------------------------------------------------------------------
# Clean up the incoming text
# -----------------------------------------------------------------------------
sub clean_text {
   my $line = shift;

   # Remove unwanted characters
   $$line =~ s/\n//g;
   $$line =~ s/\t//g;

   # Strip out HTML tags
   $$line =~ s/<.*?>//g;

   # Clean trailing "0" (aimSniff bug?)
   $$line =~ s/0$//g;
}

# -----------------------------------------------------------------------------
# Retrieve and process command line arguments
# -----------------------------------------------------------------------------
sub get_arguments {
   getopts('f:l:o:t:d:m:y:h', \%opts) or &print_usage();

   # Print help/usage information to the screen if necessary
   &print_usage() if ($opts{h});
   &print_usage() unless ($ARGV[0]);
   
   # Copy command line arguments to internal variables
   @input_files   = @ARGV;
   $filter_regexp = "." unless ($filter_regexp = $opts{f});
   $hitlist_file  = 0   unless ($hitlist_file  = $opts{l});
   $output_file   = 0   unless ($output_file   = $opts{o});
   $message_type  = 0   unless ($message_type  = $opts{t});
   $filter_day    = 0   unless ($filter_day    = $opts{d});
   $filter_month  = 0   unless ($filter_month  = $opts{m});
   $filter_year   = 0   unless ($filter_year   = $opts{y});
}

# -----------------------------------------------------------------------------
# Print usage/help information to the screen and exit
# -----------------------------------------------------------------------------
sub print_usage {
   die <<USAGE;

 +--------------------------------------------------------------+
 |           aimChomp -- an aimSniff text dump parser           |
 +--------------------------------------------------------------+

 Usage:
    perl aimchomp.pl [ -f filter ] [ -l file ] [ -o file ] 
                     [ -t type ] [ -d +|=|-day ] [ -m +|=|-month ]
                     [ -y +|=|-year ] [ file1 .. fileN ]
 Options:
    -f | Specify a regular expression filter
    -l | Output filename for a hitlist scan
    -o | Output filename; STDOUT used if not specified
    -t | Specify specific message type(s) to output
    -d | Filter (+) newer, (=) equal, (-) older then specified day
    -m | Filter (+) newer, (=) equal, (-) older then specified month
    -y | Filter (+) newer, (=) equal, (-) older then specified year

 Additional Options:
    -h | Prints this help/usage information to the screen

 Examples:
    perl aimchomp.pl -o out.txt -f poop* -t "40006" in.txt
    perl aimchomp.pl -l hitlist.txt -d +10 in1.txt in2.txt

USAGE
}
